# This file is auto-generated with util/get_builtin_functions
# for MATLAB release R2017b

HIGH_IMPACT_BUILTIN_FUNCTIONS = frozenset([
    "Inf",
    "NaN",
    "cd",
    "clear",
    "clock",
    "close",
    "cputime",
    "double",
    "eps",
    "exit",
    "eye",
    "false",
    "flintmax",
    "horzcat",
    "inf",
    "int16",
    "int32",
    "int64",
    "int8",
    "isdeployed",
    "lasterr",
    "lastwarn",
    "license",
    "more",
    "nan",
    "nargin",
    "nargout",
    "newline",
    "ones",
    "pi",
    "quit",
    "rand",
    "randn",
    "realmax",
    "realmin",
    "single",
    "speye",
    "tic",
    "toc",
    "true",
    "uint16",
    "uint32",
    "uint64",
    "uint8",
    "vertcat",
    "zeros",
])

BUILTIN_FUNCTIONS = frozenset([
    "Inf",
    "NaN",
    "abs",
    "accumarray",
    "acos",
    "acosd",
    "acosh",
    "acot",
    "acotd",
    "acoth",
    "acsc",
    "acscd",
    "acsch",
    "actxGetRunningServer",
    "actxcontrol",
    "airy",
    "all",
    "amd",
    "ancestor",
    "and",
    "animatedline",
    "ans",
    "any",
    "arith",
    "arrayfun",
    "asec",
    "asecd",
    "asech",
    "asin",
    "asind",
    "asinh",
    "assert",
    "assignin",
    "atan",
    "atan2",
    "atan2d",
    "atand",
    "atanh",
    "axes",
    "balance",
    "bandwidth",
    "beep",
    "besselh",
    "besseli",
    "besselj",
    "besselk",
    "bessely",
    "betainc",
    "betaincinv",
    "bitand",
    "bitcmp",
    "bitget",
    "bitor",
    "bitset",
    "bitshift",
    "bitxor",
    "bsxfun",
    "builtin",
    "calllib",
    "cameratoolbar",
    "cast",
    "cat",
    "cd",
    "ceil",
    "cell",
    "cell2struct",
    "cellfun",
    "changeNotification",
    "changeNotificationAdvanced",
    "char",
    "chol",
    "cholupdate",
    "circshift",
    "class",
    "clc",
    "clear",
    "clock",
    "close",
    "colamd",
    "colon",
    "colstyle",
    "complex",
    "computer",
    "conj",
    "contourc",
    "conv2",
    "convhull",
    "copyfile",
    "copyobj",
    "copyoptionsfcn",
    "cos",
    "cosd",
    "cosh",
    "cot",
    "cotd",
    "coth",
    "cputime",
    "csc",
    "cscd",
    "csch",
    "ctranspose",
    "cummax",
    "cummin",
    "cumprod",
    "cumsum",
    "dataread",
    "dbclear",
    "dbcont",
    "dbdown",
    "dbquit",
    "dbstack",
    "dbstatus",
    "dbstep",
    "dbstop",
    "dbtype",
    "dbup",
    "ddeadv",
    "ddeexec",
    "ddeinit",
    "ddepoke",
    "ddereq",
    "ddeterm",
    "ddeunadv",
    "deblank",
    "delaunay",
    "delete",
    "det",
    "diag",
    "diary",
    "diff",
    "dir",
    "disp",
    "display",
    "dissect",
    "dmperm",
    "double",
    "dragrect",
    "drawnow",
    "echo",
    "eig",
    "enumeration",
    "eps",
    "eq",
    "erf",
    "erfc",
    "erfcinv",
    "erfcx",
    "erfinv",
    "error",
    "etree",
    "eval",
    "evalc",
    "evalin",
    "events",
    "exist",
    "exit",
    "exp",
    "expm1",
    "eye",
    "false",
    "fclose",
    "feof",
    "ferror",
    "feval",
    "fft",
    "fftn",
    "fftw",
    "fgets",
    "fieldnames",
    "figure",
    "fileattrib",
    "filemarker",
    "fill",
    "fill3",
    "filter",
    "find",
    "findall",
    "findobj",
    "findstr",
    "fix",
    "flintmax",
    "flip",
    "floor",
    "fopen",
    "format",
    "fprintf",
    "frame2im",
    "fread",
    "fscanf",
    "fseek",
    "ftell",
    "full",
    "func2str",
    "functions",
    "fwrite",
    "gamma",
    "gammainc",
    "gammaincinv",
    "gammaln",
    "gcbo",
    "gcf",
    "gco",
    "ge",
    "get",
    "getappdata",
    "getenv",
    "getmcruserdata",
    "gobjects",
    "graphicsversion",
    "groot",
    "gt",
    "handle",
    "handle2struct",
    "hess",
    "hgclose",
    "hgconvertunits",
    "hgexport",
    "hggroup",
    "hgload",
    "hgtransform",
    "histc",
    "home",
    "horzcat",
    "hypot",
    "i",
    "ichol",
    "ifft",
    "ifftn",
    "ilu",
    "im2frame",
    "imag",
    "image",
    "inf",
    "inferiorto",
    "inmem",
    "input",
    "inputname",
    "int16",
    "int32",
    "int64",
    "int8",
    "inv",
    "isa",
    "isappdata",
    "isbanded",
    "iscell",
    "iscellstr",
    "ischar",
    "iscolumn",
    "isdeployed",
    "isdiag",
    "isempty",
    "isequal",
    "isequaln",
    "isequalwithequalnans",
    "isfield",
    "isfile",
    "isfinite",
    "isfloat",
    "isfolder",
    "isgraphics",
    "ishandle",
    "ishermitian",
    "ishghandle",
    "isinf",
    "isinteger",
    "isjava",
    "isletter",
    "islogical",
    "ismatrix",
    "ismcc",
    "ismembertol",
    "isnan",
    "isnumeric",
    "isplotchild",
    "isreal",
    "isrow",
    "isscalar",
    "issorted",
    "issortedrows",
    "isspace",
    "issparse",
    "isstr",
    "isstring",
    "isstrprop",
    "isstruct",
    "isstudent",
    "issymmetric",
    "istril",
    "istriu",
    "isvector",
    "j",
    "java",
    "javaArray",
    "javaMethod",
    "javaMethodEDT",
    "javaObject",
    "javaObjectEDT",
    "jpropeditutils",
    "jsondecode",
    "jsonencode",
    "keyboard",
    "lasterr",
    "lasterror",
    "lastwarn",
    "ldivide",
    "ldl",
    "le",
    "length",
    "libisloaded",
    "license",
    "light",
    "line",
    "linsolve",
    "lists",
    "load",
    "loadlibrary",
    "loadobj",
    "localfunctions",
    "log",
    "log10",
    "log1p",
    "log2",
    "logical",
    "loglog",
    "logo",
    "lookfor",
    "lower",
    "lt",
    "ltitr",
    "lu",
    "matlabpath",
    "matlabroot",
    "max",
    "maxk",
    "memory",
    "message",
    "metaclass",
    "methods",
    "mex",
    "mex_helper",
    "mexext",
    "mfilename",
    "min",
    "mink",
    "minus",
    "mislocked",
    "mkdir",
    "mldivide",
    "mlock",
    "mod",
    "more",
    "movefile",
    "movie",
    "movmad",
    "movmax",
    "movmean",
    "movmedian",
    "movmin",
    "movprod",
    "movstd",
    "movsum",
    "movvar",
    "mpower",
    "mrdivide",
    "mtimes",
    "munlock",
    "mwsamp",
    "namelengthmax",
    "nan",
    "nargchk",
    "nargin",
    "narginchk",
    "nargout",
    "nargoutchk",
    "native2unicode",
    "ndims",
    "ne",
    "newline",
    "newplot",
    "nnz",
    "nonzeros",
    "norm",
    "not",
    "numArgumentsFromSubscript",
    "numel",
    "nzmax",
    "ones",
    "or",
    "ordeig",
    "ordqz",
    "ordschur",
    "pack",
    "paren",
    "partialpath",
    "patch",
    "pause",
    "permute",
    "pi",
    "plot",
    "plot3",
    "plotyy",
    "plus",
    "polaraxes",
    "pow2",
    "power",
    "precedence",
    "prod",
    "properties",
    "psi",
    "punct",
    "py",
    "pyargs",
    "pyversion",
    "qr",
    "qrupdate",
    "quit",
    "qz",
    "rand",
    "randi",
    "randn",
    "randperm",
    "rbbox",
    "rcond",
    "rdivide",
    "real",
    "reallog",
    "realmax",
    "realmin",
    "realpow",
    "realsqrt",
    "rectangle",
    "recycle",
    "regexp",
    "regexpi",
    "regexprep",
    "regexptranslate",
    "rehash",
    "relop",
    "rem",
    "repelem",
    "repmat",
    "reset",
    "reshape",
    "rethrow",
    "rmappdata",
    "rmdir",
    "round",
    "save",
    "savefig",
    "saveobj",
    "schur",
    "script",
    "sec",
    "secd",
    "sech",
    "selectmoveresize",
    "semilogx",
    "semilogy",
    "seriallist",
    "set",
    "setappdata",
    "setenv",
    "setmcruserdata",
    "setstr",
    "setupplotbrowser",
    "sign",
    "sin",
    "sind",
    "single",
    "sinh",
    "size",
    "slash",
    "sort",
    "sparse",
    "speye",
    "sprintf",
    "sqrt",
    "sscanf",
    "str2func",
    "strcmp",
    "strcmpi",
    "strfind",
    "strings",
    "strncmp",
    "strncmpi",
    "strrep",
    "strtrim",
    "struct",
    "struct2cell",
    "struct2handle",
    "structfun",
    "subsasgn",
    "subsindex",
    "subsref",
    "sum",
    "superclasses",
    "superiorto",
    "support",
    "surface",
    "svd",
    "symamd",
    "symbfact",
    "symrcm",
    "syntax",
    "system",
    "tan",
    "tand",
    "tanh",
    "tex",
    "text",
    "textscan",
    "tic",
    "times",
    "toc",
    "transpose",
    "tril",
    "triu",
    "true",
    "type",
    "typecast",
    "uicontainer",
    "uicontextmenu",
    "uicontrol",
    "uiflowcontainer",
    "uigetdir",
    "uigetfile",
    "uigridcontainer",
    "uimenu",
    "uint16",
    "uint32",
    "uint64",
    "uint8",
    "uipanel",
    "uipushtool",
    "uiputfile",
    "uisetcolor",
    "uisetfont",
    "uitable",
    "uitoggletool",
    "uitoolbar",
    "uminus",
    "unicode2native",
    "uniquetol",
    "unloadlibrary",
    "uplus",
    "upper",
    "varargin",
    "varargout",
    "vecnorm",
    "vertcat",
    "vissuite",
    "waitfor",
    "waitforbuttonpress",
    "warning",
    "what",
    "which",
    "who",
    "whos",
    "winqueryreg",
    "xor",
    "yyaxis",
    "zeros",
    "zoom",
])

EXTRA_FUNCTIONS = frozenset([
    "NaT",
    "activateuimode",
    "actxcontrolcreateproperty",
    "actxcontrollist",
    "actxcontrolselect",
    "actxproxy",
    "actxserver",
    "adddatadlg",
    "addpath",
    "addpref",
    "addsubplot",
    "addtodate",
    "align",
    "alim",
    "allMuPADNotebooks",
    "allchild",
    "alpha",
    "alphamap",
    "analyzeCodeCompatibility",
    "angle",
    "annotation",
    "appdesigner",
    "applicationCompiler",
    "arduino",
    "area",
    "array2table",
    "array2timetable",
    "arrayviewfunc",
    "audiodevinfo",
    "audioinfo",
    "audioread",
    "audiouniquename",
    "audiowrite",
    "auditcontents",
    "automesh",
    "autumn",
    "avgate",
    "avifinfo",
    "aviinfo",
    "awtcreate",
    "awtinvoke",
    "axescheck",
    "axis",
    "bar",
    "bar3",
    "bar3h",
    "barh",
    "base2dec",
    "basicfitdatastat",
    "bench",
    "beta",
    "betaln",
    "bicg",
    "bicgstab",
    "bicgstabl",
    "bin2dec",
    "binscatter",
    "blanks",
    "blkdiag",
    "bone",
    "boundary",
    "bounds",
    "box",
    "brighten",
    "brush",
    "btndown",
    "btngroup",
    "btnicon",
    "btnpress",
    "btnresize",
    "btnstate",
    "btnup",
    "builddocsearchdb",
    "bvp4c",
    "bvp5c",
    "bvpget",
    "bvpinit",
    "bvpset",
    "bvpxtend",
    "bwcontr",
    "caldays",
    "calendar",
    "callAllOptimOutputFcns",
    "callAllOptimPlotFcns",
    "callSoapService",
    "callgraphviz",
    "calmonths",
    "calquarters",
    "calweeks",
    "calyears",
    "camdolly",
    "cameraCalibrator",
    "cameramenu",
    "cameratoolbar",
    "camlight",
    "camlookat",
    "camorbit",
    "campan",
    "campos",
    "camproj",
    "camroll",
    "camrotate",
    "camtarget",
    "camup",
    "camva",
    "camzoom",
    "capturescreen",
    "cart2pol",
    "cart2sph",
    "caxis",
    "cdf2rdf",
    "cdfinfo",
    "cdfread",
    "cdfwrite",
    "cell2mat",
    "cell2table",
    "celldisp",
    "cellplot",
    "cellstr",
    "cftool",
    "cgs",
    "changeseriestype",
    "checkSyntacticWarnings",
    "checkcode",
    "checkin",
    "checkout",
    "cla",
    "clabel",
    "classificationLearner",
    "clearAllMemoizedCaches",
    "clearvars",
    "clf",
    "clipboard",
    "closereq",
    "cmapeditor",
    "cmopts",
    "cmpermute",
    "cmunique",
    "code2html",
    "codeCompatibilityReport",
    "codegen",
    "codetoolsswitchyard",
    "colorThresholder",
    "colorbar",
    "colorcube",
    "colordef",
    "colormap",
    "colormapeditor",
    "colornone",
    "colorpalette",
    "colperm",
    "comet",
    "comet3",
    "commandhistory",
    "commandwindow",
    "commonplotfunc",
    "compan",
    "compass",
    "compose",
    "cond",
    "condeig",
    "condest",
    "coneplot",
    "connector",
    "consume_assign",
    "contains",
    "contentsrpt",
    "contour",
    "contour3",
    "contourf",
    "contours",
    "contourslice",
    "contrast",
    "controlSystemTuner",
    "conv",
    "convertCharsToStrings",
    "convertSpreadsheetDates",
    "convertSpreadsheetExcelDates",
    "convertStringsToChars",
    "convhulln",
    "convn",
    "cool",
    "copper",
    "corrcoef",
    "count",
    "cov",
    "coveragerpt",
    "cplxpair",
    "createCellArrayOfFunctions",
    "createClassFromWsdl",
    "createProblemStruct",
    "createSoapMessage",
    "createSolverStruct",
    "cross",
    "cshelp",
    "csvread",
    "csvwrite",
    "cumtrapz",
    "curl",
    "customverctrl",
    "cylinder",
    "daexplr",
    "daqread",
    "daspect",
    "datachildren",
    "datacursormode",
    "datastore",
    "datatipinfo",
    "date",
    "dateTickPicker",
    "datenum",
    "datestr",
    "datetick",
    "datetickstr",
    "datevec",
    "days",
    "dblquad",
    "dbmex",
    "dde23",
    "ddeget",
    "ddensd",
    "ddesd",
    "ddeset",
    "deal",
    "dec2base",
    "dec2bin",
    "dec2hex",
    "decic",
    "deconv",
    "defaultItemHitCallback",
    "defaulterrorcallback",
    "deg2rad",
    "del2",
    "delaunayn",
    "delsq",
    "demo",
    "demos",
    "demowin",
    "depfunprophelper",
    "deployprefdir",
    "deploytool",
    "deprpt",
    "deselectall",
    "desktop",
    "desktopmenufcn",
    "details",
    "detrend",
    "deval",
    "dialog",
    "diary",
    "diffuse",
    "discretize",
    "displayEndOfDemoMessage",
    "dither",
    "divergence",
    "dlmread",
    "dlmwrite",
    "doc",
    "doclick",
    "doclink",
    "docpath",
    "docroot",
    "docsearch",
    "dofixrpt",
    "dokeypress",
    "domymenu",
    "doresize",
    "dos",
    "dot",
    "dsearchn",
    "dspfwiz",
    "echodemo",
    "edit",
    "editmenufcn",
    "eigs",
    "ellipj",
    "ellipke",
    "ellipsoid",
    "enableNETfromNetworkDrive",
    "enableplottoolbuttons",
    "enableservice",
    "enddrag",
    "endsWith",
    "eomday",
    "erase",
    "eraseBetween",
    "errorbar",
    "errordlg",
    "etime",
    "etreeplot",
    "eventlisteners",
    "exit",
    "expint",
    "expm",
    "export2wsdlg",
    "exportsetupdlg",
    "extractAfter",
    "extractBefore",
    "extractBetween",
    "ezcontour",
    "ezcontourf",
    "ezgraph3",
    "ezmesh",
    "ezmeshc",
    "ezplot",
    "ezplot3",
    "ezpolar",
    "ezsurf",
    "ezsurfc",
    "factor",
    "factorial",
    "fcnchk",
    "fcontour",
    "feather",
    "fft2",
    "fftshift",
    "fgetl",
    "figcopytemplatelistener",
    "figcopytemplatelistener_legacy",
    "figflag",
    "fignamer",
    "figtoolset",
    "figureToolbarCreateFcn",
    "figuredeletedlistener",
    "figureheaderdlg",
    "figurepalette",
    "fileDatastore",
    "filebrowser",
    "filemenufcn",
    "fileparts",
    "fileread",
    "filesep",
    "fillmissing",
    "filloutliers",
    "filter2",
    "filterBuilder",
    "fimplicit",
    "fimplicit3",
    "findExample",
    "findall",
    "findfigs",
    "findgroups",
    "findprinters",
    "finfo",
    "fitsdisp",
    "fitsinfo",
    "fitsread",
    "fitswrite",
    "fixcontents",
    "fixquote",
    "flag",
    "flipdim",
    "fliplr",
    "flipud",
    "fmesh",
    "fminbnd",
    "fminsearch",
    "fplot",
    "fplot3",
    "freqspace",
    "frewind",
    "fsurf",
    "fullfile",
    "function_handle",
    "functionhintsfunc",
    "functiontests",
    "funm",
    "fzero",
    "gallery",
    "gather",
    "gca",
    "gcbf",
    "gcd",
    "gcmr",
    "generateProfilerSelect",
    "genpath",
    "genvarname",
    "geobubble",
    "geolimits",
    "getFigureToolManager",
    "getcallinfo",
    "getcolumn",
    "getcrosssection",
    "getfield",
    "getfigurefordesktopclient",
    "getframe",
    "getobj",
    "getorcreateobj",
    "getpixelposition",
    "getplotbrowserproptable",
    "getplottool",
    "getpref",
    "getprinttemplate",
    "getptr",
    "getrangefromclass",
    "getscribecontextmenu",
    "getscribeobjectdata",
    "getselectobjects",
    "getstatus",
    "getuimode",
    "ginput",
    "gmres",
    "gobjects",
    "gplot",
    "grabcode",
    "gradient",
    "graph2dhelper",
    "gray",
    "graymon",
    "grid",
    "griddata",
    "griddatan",
    "gsvd",
    "gtext",
    "gui_mainfcn",
    "guidata",
    "guide",
    "guidefunc",
    "guidemfile",
    "guideopts",
    "guidetemplate",
    "guidetemplate0",
    "guidetemplate1",
    "guidetemplate2",
    "guidetemplate3",
    "guidetoolfunc",
    "guihandles",
    "gunzip",
    "gzip",
    "h5create",
    "h5disp",
    "h5info",
    "h5read",
    "h5readatt",
    "h5write",
    "h5writeatt",
    "hadamard",
    "hankel",
    "hardcopy",
    "hasbehavior",
    "hasuimode",
    "hdf",
    "hdf5",
    "hdf5info",
    "hdf5read",
    "hdf5write",
    "hdfan",
    "hdfdf24",
    "hdfdfr8",
    "hdfgd",
    "hdfh",
    "hdfhd",
    "hdfhe",
    "hdfhx",
    "hdfinfo",
    "hdfml",
    "hdfpt",
    "hdfread",
    "hdfsd",
    "hdfsw",
    "hdftool",
    "hdfv",
    "hdfvf",
    "hdfvh",
    "hdfvs",
    "heatmap",
    "help",
    "help2html",
    "helpPopup",
    "helpbrowser",
    "helpdesk",
    "helpdlg",
    "helpmenufcn",
    "helprpt",
    "helpview",
    "helpwin",
    "hex2dec",
    "hex2num",
    "hgaddbehavior",
    "hgbehaviorfactory",
    "hgdisp",
    "hgfeval",
    "hggetbehavior",
    "hggetdisp",
    "hgline2lineseries",
    "hgload",
    "hgsave",
    "hgsetdisp",
    "hidden",
    "hilb",
    "hist",
    "histcounts",
    "histcounts2",
    "histogram",
    "histogram2",
    "hold",
    "home",
    "hot",
    "hours",
    "hsv",
    "hsv2rgb",
    "icondisp",
    "iconeditor",
    "iconread",
    "idivide",
    "ifft2",
    "ifftshift",
    "im2double",
    "im2java",
    "imageBatchProcessor",
    "imageDatastore",
    "imageRegionAnalyzer",
    "imageSegmenter",
    "imagesc",
    "imageview",
    "imapprox",
    "imfinfo",
    "imformats",
    "implay",
    "importdata",
    "imread",
    "imresize",
    "imshow",
    "imwrite",
    "ind2rgb",
    "ind2sub",
    "indentcode",
    "indexhelper",
    "info",
    "initdesktoputils",
    "inlineeval",
    "inpolygon",
    "input",
    "inputdlg",
    "insertAfter",
    "insertBefore",
    "insertmenufcn",
    "inspect",
    "instrcb",
    "instrfind",
    "instrfindall",
    "int2str",
    "integral",
    "integral2",
    "integral3",
    "interp1",
    "interp1q",
    "interp2",
    "interp3",
    "interpft",
    "interpn",
    "interpstreamspeed",
    "intersect",
    "intmax",
    "intmin",
    "invhilb",
    "invokeProfiler",
    "ipcam",
    "ipermute",
    "is2D",
    "isStringScalar",
    "isactiveuimode",
    "iscalendarduration",
    "iscategorical",
    "ischange",
    "iscom",
    "isdatetime",
    "isdicom",
    "isdir",
    "isdpx",
    "isduration",
    "isenum",
    "isevent",
    "ishold",
    "isinterface",
    "iskeyword",
    "islocalmax",
    "islocalmin",
    "ismac",
    "ismember",
    "ismethod",
    "ismissing",
    "isnitf",
    "isobject",
    "isocaps",
    "isocolors",
    "isonormals",
    "isosurface",
    "isoutlier",
    "ispc",
    "isplotchild",
    "ispref",
    "isprime",
    "isprop",
    "istable",
    "istall",
    "istimetable",
    "isunix",
    "isvarname",
    "javaAddLsnrsToFigure",
    "javaGetHandles",
    "javaaddpath",
    "javachk",
    "javaclasspath",
    "javacomponent",
    "javarmpath",
    "jet",
    "join",
    "keyboard",
    "kron",
    "lcm",
    "legend",
    "legendcolorbarlayout",
    "legendinfo",
    "legendre",
    "libfunctions",
    "libfunctionsview",
    "libpointer",
    "libstruct",
    "lightangle",
    "lighting",
    "lin2mu",
    "lines",
    "linkaxes",
    "linkdata",
    "linkplotfunc",
    "linkprop",
    "linspace",
    "listdlg",
    "listfonts",
    "loadlibrary",
    "logm",
    "logspace",
    "ls",
    "lscov",
    "lsqminnorm",
    "lsqnonneg",
    "lsqr",
    "m2struct",
    "magic",
    "makePixelPath",
    "makebars",
    "makecontentsfile",
    "makedisplaynames",
    "makehgtform",
    "makemcode",
    "makemenu",
    "makeprofilerfooter",
    "makeprofilerheader",
    "mapreduce",
    "mapreducer",
    "mat2cell",
    "mat2str",
    "material",
    "matfile",
    "matfinfo",
    "maxNumCompThreads",
    "mdbfileonpath",
    "mdbpublish",
    "mdbstatus",
    "mean",
    "median",
    "membrane",
    "memoize",
    "menu",
    "menueditfunc",
    "menulabel",
    "mesh",
    "meshc",
    "meshgrid",
    "meshz",
    "methodsview",
    "mex",
    "mex_legacy",
    "mexdebug",
    "middrag",
    "milliseconds",
    "minres",
    "minutes",
    "mkpp",
    "mlappfinfo",
    "mlappinstallfinfo",
    "mlint",
    "mlintrpt",
    "mltbxfinfo",
    "mltimerpackage",
    "mlxdoc",
    "mmfileinfo",
    "mobiledev",
    "mode",
    "more",
    "moveaxis",
    "movegui",
    "movie2avi",
    "moviein",
    "movieview",
    "msgbox",
    "mu2lin",
    "multibandread",
    "multibandwrite",
    "mupadwelcome",
    "mustBeFinite",
    "mustBeGreaterThan",
    "mustBeGreaterThanOrEqual",
    "mustBeInteger",
    "mustBeLessThan",
    "mustBeLessThanOrEqual",
    "mustBeMember",
    "mustBeNegative",
    "mustBeNonNan",
    "mustBeNonempty",
    "mustBeNonnegative",
    "mustBeNonpositive",
    "mustBeNonsparse",
    "mustBeNonzero",
    "mustBeNumeric",
    "mustBeNumericOrLogical",
    "mustBePositive",
    "mustBeReal",
    "nccreate",
    "ncdisp",
    "nchoosek",
    "ncinfo",
    "ncread",
    "ncreadatt",
    "ncwrite",
    "ncwriteatt",
    "ncwriteschema",
    "ndgrid",
    "nested",
    "nextpow2",
    "noanimate",
    "nodither",
    "normest",
    "normest1",
    "now",
    "nsidedpoly",
    "nthroot",
    "null",
    "num2cell",
    "num2hex",
    "num2ruler",
    "num2str",
    "numgrid",
    "numjac",
    "objbounds",
    "ocrTrainer",
    "ode113",
    "ode15i",
    "ode15s",
    "ode23",
    "ode23s",
    "ode23t",
    "ode23tb",
    "ode45",
    "odeexamples",
    "odeget",
    "odephas2",
    "odephas3",
    "odeplot",
    "odeprint",
    "odeset",
    "odextend",
    "online_concatenator",
    "open",
    "openExample",
    "openFigure",
    "open_system",
    "openfig",
    "opengl",
    "openmlapp",
    "openmn",
    "openmu",
    "opentoline",
    "openvar",
    "openxvc",
    "openxvz",
    "optimget",
    "optimplotfunccount",
    "optimplotfval",
    "optimplotx",
    "optimset",
    "orderfields",
    "orient",
    "orth",
    "overobj",
    "pad",
    "pagesetupdlg",
    "pan",
    "parallel_function",
    "pareto",
    "parfor_M_check",
    "parfor_endpoint_check",
    "parfor_range_check",
    "parfor_sliced_fcnhdl_check",
    "parseSoapResponse",
    "parseparams",
    "parseplotapi",
    "parula",
    "pascal",
    "path",
    "path2rc",
    "pathsep",
    "pathtool",
    "pause",
    "pbaspect",
    "pcg",
    "pchip",
    "pcode",
    "pcolor",
    "pdepe",
    "pdeval",
    "peaks",
    "perl",
    "perms",
    "pie",
    "pie3",
    "pink",
    "pinv",
    "planerot",
    "playshow",
    "plotbrowser",
    "plotchild",
    "plotdoneevent",
    "plotedit",
    "plotmatrix",
    "plotpickerfunc",
    "plottoolfunc",
    "plottools",
    "plotutils",
    "plotyy",
    "pol2cart",
    "polar",
    "polarhistogram",
    "polarplot",
    "polarscatter",
    "poly",
    "polyarea",
    "polyder",
    "polyeig",
    "polyfit",
    "polyint",
    "polyspaceBugFinder",
    "polyspaceCodeProver",
    "polyval",
    "polyvalm",
    "popupstr",
    "ppval",
    "prefdir",
    "preferences",
    "prefspanel",
    "prepdrag",
    "primes",
    "print",
    "printProfilerRefreshButton",
    "printdlg",
    "printfProfilerLink",
    "printjob",
    "printpreview",
    "printtables",
    "printtemplate",
    "prism",
    "profile",
    "profreport",
    "profsave",
    "profview",
    "profviewgateway",
    "projdumpmat",
    "propedit",
    "propertyeditor",
    "publish",
    "putdowntext",
    "pwch",
    "pwd",
    "qhull",
    "qmr",
    "qrdelete",
    "qrinsert",
    "quad",
    "quad2d",
    "quadgk",
    "quadl",
    "quadv",
    "questdlg",
    "quit",
    "quiver",
    "quiver3",
    "rad2deg",
    "rank",
    "raspberrypi",
    "raspi",
    "rat",
    "rats",
    "readMobileSensorData",
    "readtable",
    "rectint",
    "reducepatch",
    "reducevolume",
    "refresh",
    "refreshdata",
    "regionpicker",
    "registerevent",
    "regmatlabserver",
    "remapfig",
    "renameStructField",
    "replace",
    "replaceBetween",
    "rescale",
    "resetplotview",
    "reverse",
    "reverse_binary_operator",
    "rfBudgetAnalyzer",
    "rgb2gray",
    "rgb2hsv",
    "rgb2ind",
    "rgbplot",
    "ribbon",
    "rjr",
    "rlim",
    "rmfield",
    "rmmissing",
    "rmpath",
    "rmpref",
    "rng",
    "roots",
    "rose",
    "rosser",
    "rot90",
    "rotate",
    "rotate3d",
    "rref",
    "rsf2csf",
    "rtickangle",
    "rtickformat",
    "rticklabels",
    "rticks",
    "ruler2num",
    "run",
    "runperf",
    "runreport",
    "runtests",
    "sampev",
    "saveFigure",
    "saveas",
    "savefig",
    "savepath",
    "savtoner",
    "scatter",
    "scatter3",
    "scribealign",
    "scribeccp",
    "scribeclearmode",
    "scribecursors",
    "scribeeventhandler",
    "scriberestoresavefcns",
    "scribetextdlg",
    "seconds",
    "selectobject",
    "sendmail",
    "separateOptimStruct",
    "setProfilerHtmlText",
    "setdiff",
    "setfield",
    "setpixelposition",
    "setpref",
    "setprinttemplate",
    "setptr",
    "setscribecontextmenu",
    "setscribeobjectdata",
    "setstatus",
    "settings",
    "setupExample",
    "setxor",
    "shading",
    "sharedplotfunc",
    "shg",
    "shiftdim",
    "showAddonExplorer",
    "showAddonExplorerForHardwareSupportPackages",
    "showHelpPreferences",
    "showdemo",
    "showplottool",
    "shrinkfaces",
    "signalAnalyzer",
    "simbiology",
    "simulink",
    "simulinkts2ts",
    "slLibraryBrowser",
    "slice",
    "smooth3",
    "smoothdata",
    "snapnow",
    "snaptogrid",
    "snlsFixedVarEvalOutputOrPlotFcn",
    "snlsFixedVarWrapOutputAndPlotFcns",
    "sortrows",
    "sound",
    "soundsc",
    "soundview",
    "spalloc",
    "spaugment",
    "spconvert",
    "spdiags",
    "specgraphhelper",
    "specular",
    "spfun",
    "sph2cart",
    "sphere",
    "spinmap",
    "spline",
    "split",
    "splitapply",
    "splitlines",
    "splncore",
    "spmd_feval",
    "spones",
    "spparms",
    "sprand",
    "sprandn",
    "sprandsym",
    "sprank",
    "spreadsheetDatastore",
    "spring",
    "spy",
    "sqrtm",
    "squeeze",
    "stairs",
    "standardizeMissing",
    "startsWith",
    "startscribeobject",
    "startscribepinning",
    "stateflow",
    "std",
    "stem",
    "stem3",
    "stereoCameraCalibrator",
    "str2double",
    "str2mat",
    "str2num",
    "strcat",
    "stream2",
    "stream3",
    "streamline",
    "streamparticles",
    "streamribbon",
    "streamslice",
    "streamtube",
    "strip",
    "stripanchors",
    "strjoin",
    "strjust",
    "strlength",
    "strmatch",
    "strread",
    "strsplit",
    "strtok",
    "struct2table",
    "strvcat",
    "sub2ind",
    "subplot",
    "subspace",
    "substruct",
    "subvolume",
    "summer",
    "superiorfloat",
    "support",
    "supportPackageInstaller",
    "surf",
    "surf2patch",
    "surfc",
    "surfl",
    "surfnorm",
    "svds",
    "swapbytes",
    "sylvester",
    "symmlq",
    "symvar",
    "tabdlg",
    "table2array",
    "table2cell",
    "table2struct",
    "table2timetable",
    "tabularTextDatastore",
    "tallrng",
    "tar",
    "tempdir",
    "tempname",
    "testsuite",
    "tetramesh",
    "texlabel",
    "textread",
    "textwrap",
    "tfqmr",
    "thetalim",
    "thetatickformat",
    "thetaticklabels",
    "thetaticks",
    "thingSpeakArea",
    "thingSpeakAuthenticate",
    "thingSpeakAuthenticatedList",
    "thingSpeakClearAuthentication",
    "thingSpeakPlot",
    "thingSpeakPlotYY",
    "thingSpeakRead",
    "thingSpeakScatter",
    "thingSpeakStem",
    "thingSpeakWrite",
    "timeit",
    "timercb",
    "timerfind",
    "timerfindall",
    "timetable2table",
    "timezones",
    "tipoftheday",
    "title",
    "toeplitz",
    "toolbareditor",
    "toolboxdir",
    "toolpalette",
    "toolsmenufcn",
    "topkrows",
    "trace",
    "trainingImageLabeler",
    "trapz",
    "treelayout",
    "treeplot",
    "trimesh",
    "triplequad",
    "triplot",
    "trisurf",
    "tsAlignSizes",
    "tsAnalyzeAbsTime",
    "tsChkTime",
    "tsIsDateFormat",
    "tsIsSameTime",
    "tsarrayFcn",
    "tsdateinterval",
    "tsearchn",
    "tsfcncallrouter",
    "tsgetDateFormat",
    "tsgetrelativetime",
    "tsinterp",
    "tsnaniqr",
    "tsnanmax",
    "tsnanmean",
    "tsnanmedian",
    "tsnanmin",
    "tsnanmode",
    "tsnanstd",
    "tsnansum",
    "tsnanvar",
    "tsnewevent",
    "tspnmatch",
    "tsprctile",
    "tspvformat",
    "tssorttime",
    "tsunitconv",
    "uialert",
    "uiaxes",
    "uibutton",
    "uibuttongroup",
    "uicheckbox",
    "uiclearmode",
    "uiconfirm",
    "uicontrol",
    "uidropdown",
    "uieditfield",
    "uifigure",
    "uigauge",
    "uigetdir",
    "uigetfile",
    "uigetmodemanager",
    "uigetpref",
    "uigettool",
    "uigettoolbar",
    "uiimport",
    "uiknob",
    "uilabel",
    "uilamp",
    "uilistbox",
    "uiload",
    "uimode",
    "uiopen",
    "uiputfile",
    "uiradiobutton",
    "uirestore",
    "uiresume",
    "uisave",
    "uisetcolor",
    "uisetfont",
    "uisetpref",
    "uislider",
    "uispinner",
    "uistack",
    "uisuspend",
    "uiswitch",
    "uitab",
    "uitabgroup",
    "uitextarea",
    "uitogglebutton",
    "uitoolfactory",
    "uitree",
    "uitreenode",
    "uiundo",
    "uiwait",
    "undocheckout",
    "union",
    "unique",
    "unix",
    "unmesh",
    "unmkpp",
    "unregisterallevents",
    "unregisterevent",
    "untar",
    "unwrap",
    "unzip",
    "updateHelpPopup",
    "urldecode",
    "urlencode",
    "urlfilter",
    "urlread",
    "urlwrite",
    "usejava",
    "usejavacomponent",
    "usev6plotapi",
    "utTrimNans",
    "uttsget",
    "validOptimProblemStruct",
    "validateattributes",
    "validateopts_UseParallel",
    "validatestring",
    "value2CustomFcn",
    "vander",
    "var",
    "vectorize",
    "ver",
    "verLessThan",
    "verctrl",
    "vertexpicker",
    "vga",
    "view",
    "viewmenufcn",
    "viewmtx",
    "volumebounds",
    "voronoi",
    "voronoin",
    "waitbar",
    "warndlg",
    "waterfall",
    "web",
    "webcam",
    "webcamlist",
    "webmap",
    "webread",
    "websave",
    "webwrite",
    "weekday",
    "whatsnew",
    "white",
    "whitebg",
    "wilkinson",
    "winmenu",
    "winopen",
    "winter",
    "wizard",
    "wordcloud",
    "workspace",
    "workspacefunc",
    "writeactxlicense",
    "writetable",
    "xlabel",
    "xlim",
    "xlsfinfo",
    "xlsread",
    "xlswrite",
    "xmlread",
    "xmlwrite",
    "xslt",
    "xtickangle",
    "xtickformat",
    "xticklabels",
    "xticks",
    "xychk",
    "xyzchk",
    "xyzvchk",
    "years",
    "ylabel",
    "ylim",
    "ytickangle",
    "ytickformat",
    "yticklabels",
    "yticks",
    "zip",
    "zlabel",
    "zlim",
    "zoom",
    "ztickangle",
    "ztickformat",
    "zticklabels",
    "zticks",
])

BUILTIN_CLASSES = frozenset([
    "COM",
    "CodeCompatibilityAnalysis",
    "DataTipUtilities",
    "DelaunayTri",
    "MException",
    "ParallelException",
    "RandStream",
    "SimTimeseries",
    "Tiff",
    "TriRep",
    "TriScatteredInterp",
    "VideoReader",
    "VideoWriter",
    "alphaShape",
    "arduino",
    "arrowline",
    "audioplayer",
    "audiorecorder",
    "axischild",
    "axisobj",
    "axistext",
    "calendarDuration",
    "categorical",
    "cdfepoch",
    "cell",
    "datetime",
    "decomposition",
    "delaunayTriangulation",
    "digraph",
    "duration",
    "dynamicprops",
    "editline",
    "editrect",
    "fighandle",
    "figobj",
    "ftp",
    "graph",
    "graphics",
    "griddedInterpolant",
    "handle",
    "hdf5",
    "hdftool",
    "hds",
    "hgbin",
    "inline",
    "inputParser",
    "internalArrayUtil",
    "memmap_data_handle_holder",
    "memmapfile",
    "missing",
    "mtree",
    "mupad",
    "onCleanup",
    "opaque",
    "polyshape",
    "printtemplate",
    "scatteredInterpolant",
    "scribe",
    "scribehandle",
    "scribehgobj",
    "serial",
    "string",
    "system",
    "table",
    "tabular",
    "tall",
    "tcpclient",
    "timer",
    "timerange",
    "timeseries",
    "timetable",
    "triangulation",
    "tscollection",
    "tsdata",
    "uitools",
    "vartype",
    "weboptions",
    "withtol",
])

BUILTIN_NAMESPACES = frozenset([
    "H5",
    "H5A",
    "H5D",
    "H5DS",
    "H5E",
    "H5F",
    "H5G",
    "H5I",
    "H5L",
    "H5ML",
    "H5O",
    "H5P",
    "H5R",
    "H5S",
    "H5T",
    "H5Z",
    "NET",
    "appcreate",
    "appdesigner",
    "appinstall",
    "audiovideo",
    "brushing",
    "cdflib",
    "cftool",
    "codegen",
    "connector",
    "containers",
    "datamanager",
    "dsp",
    "event",
    "exampleUtils",
    "help",
    "helpUtils",
    "images",
    "inspector",
    "internal",
    "matlab",
    "message",
    "meta",
    "mex",
    "mex_legacy",
    "mls",
    "mtree",
    "netcdf",
    "nnet",
    "objutil",
    "peermodel",
    "python",
    "spf",
    "thingspeak",
    "tsdata",
    "tsguis",
    "vision",
    "viz",
    "webutils",
])

SPECIAL_NAMES = frozenset([
    "Contents",
])
