body {
    background-color: #f6f6f6;
    color: #000;
    font-family: sans;
    margin: 0;
}

main {
    margin-left: 2em;
    margin-right: 2em;
}

header {
    padding: 0.5em 0.5em 0.5em 1.4em;
    background-color: cornflowerblue;
    color: white;
    font-size: 150%;
    font-weight: bold;
    box-shadow: 0 4px 10px -2px rgba(0,0,0,.2);
}

footer {
    padding: 0.5em 0.5em 0.5em 1.4em;
    color: #444;
    background-color: white;
    border-radius: 0.2em;
    padding: 0.8em;
    line-height: 1.2;
    text-align: center;
    margin-top: 1em;
}

h1 {
    padding-left: 0.2em;
    text-transform: uppercase;
    font-size: 120%;
}

.title {
    display: flex;
    align-items: center;
    padding-left: 0.4em;
}

.title img {
    width: 1.2em;
    height: 1.2em;
    stroke: cornflowerblue;
}

h2 {
    color: cornflowerblue;
    margin-block-start: 0;
    font-size: 120%;
}

h2 a:hover {
    background-color: white;
    border-radius: 0px;
}

h3 {
    color: black;
}

h3 a {
    color: black;
}

h3 a:hover {
    background-color: white;
    border-radius: 0px;
}

h4 {
    color: black;
    border-left: 0.3em solid cornflowerblue;
    padding-left: 0.2em;
}

h4 a {
    color: black;
}

h4 a:hover {
    background-color: white;
    border-radius: 0px;
}

a {
    color: cornflowerblue;
    font-weight: bold;
    text-decoration: none;
}

a:hover {
    background-color: #eef;
    border-radius: 3px;
}

span.file {
    font-family: mono;
}

div {
    margin-top: 1em;
}

div.navigation {
    text-align: center;
    font-size: 110%;
}

.navigation span.current {
    font-weight: bold;
}

.navigation span::before {
    content: "[";
    margin-left: 1em;
}

.navigation span::after {
    content: "]";
    margin-right: 1em;
}

section {
    background-color: white;
    border-radius: 0.2em;
    padding: 0.8em;
    box-shadow: 0 2px 10px 0 rgba(0,0,0,.05);
    line-height: 1.2;
    color: #444;
    margin-bottom: 0.4em;
}

section pre {
    border-radius: 0.2em;
    padding: 0.8em;
    border: 0.1em solid lightgray;
    background-color: #f6f6f6;
    line-height: normal;
    color: black;
}

.metrics h2 {
    margin-bottom: 0;
}

.metrics td {
    padding: 0.3em;
}

.metrics td:first-child {
     min-width: 12em;
}

.metrics table{
    border-radius: 0.2em;
    border: 0.1em solid lightgray;
    margin-bottom: 1.5em;
}

thead {
    background-color: #f6f6f6;
    font-weight: bold;
}

td.na {
    background-color: #f6f6f6;
}

td.ok {
    background-color: #cfc;
}

td.ok_justified {
    background-color: #ccf;
}

td.nok {
    background-color: #fcc;
}

.tip {
    position:relative;
}

.tip:hover{
    z-index:1000;
    position:relative;
}

.tip:hover:after{
    display:block;
}

.tip:hover:before{
    display:block;
}

.tip:hover{
    text-decoration:none;
}

.tip:after{
    content: attr(tip);
    position: absolute;
    padding: 0.5em;
    margin: 0.2em;
    display:none;
    background-color: white;
    top: 1em;
    left: 2.5em;
    border-radius: 0.2em;
    border: 0.1em solid lightgray;
    width: 25em;
    box-shadow: 3px 3px 4px rgba(0,0,0, .65);
}
